import sys
from pathlib import Path
from typing import Callable, Optional

import typer
from typer import Argument, Option

import tamp

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False, add_completion=False)


def version_callback(value: bool):
    if not value:
        return
    print(tamp.__version__)
    raise typer.Exit()


@app.callback()
def common(
    version: bool = Option(
        None,
        "--version",
        "-v",
        callback=version_callback,
        help="Print tamp version.",
    ),
):
    pass


def read(input_: Optional[Path]) -> bytes:
    data = sys.stdin.buffer.read() if input_ is None else input_.read_bytes()
    if not data:
        raise ValueError("No data provided.")
    return data


def write(output: Optional[Path], data: bytes):
    if output is None:
        sys.stdout.buffer.write(data)
    else:
        output.write_bytes(data)


@app.command()
def compress(
    input_path: Optional[Path] = Argument(
        None,
        exists=True,
        readable=True,
        dir_okay=False,
        show_default=False,
        help="Input file to compress or decompress. Defaults to stdin.",
    ),
    output_path: Optional[Path] = Option(
        None,
        "--output",
        "-o",
        exists=False,
        writable=True,
        dir_okay=True,
        show_default=False,
        help="Output file. Defaults to stdout.",
    ),
    window: int = Option(
        10,
        "-w",
        "--window",
        min=8,
        max=15,
        help="Number of bits used to represent the dictionary window.",
    ),
    literal: int = Option(
        8,
        "-l",
        "--literal",
        min=5,
        max=8,
        help="Number of bits used to represent a literal.",
    ),
):
    """Compress an input file or stream."""
    input_bytes = read(input_path)
    output_bytes = tamp.compress(
        input_bytes,
        window=window,
        literal=literal,
    )
    write(output_path, output_bytes)


@app.command()
def decompress(
    input_path: Optional[Path] = Argument(
        None,
        exists=True,
        readable=True,
        dir_okay=False,
        show_default=False,
        help="Input file. If not provided, reads from stdin.",
    ),
    output_path: Optional[Path] = Option(
        None,
        "--output",
        "-o",
        exists=False,
        writable=True,
        dir_okay=True,
        show_default=False,
        help="Output file. Defaults to stdout.",
    ),
):
    """Decompress an input file or stream."""
    input_bytes = read(input_path)
    output_bytes = tamp.decompress(input_bytes)
    write(output_path, output_bytes)


def run_app(*args, **kwargs):
    app(*args, **kwargs)



# Don't manually change, let poetry-dynamic-versioning-plugin handle it.
__version__ = "1.1.2"


class ExcessBitsError(Exception):
    """Provided data has more bits than expected ``literal`` bits."""


def bit_size(value):
    for i in range(32):
        if not value:
            return i
        value >>= 1
    return -1


def _xorshift32(seed):
    while True:
        seed ^= (seed << 13) & 0xFFFFFFFF
        seed ^= (seed >> 17) & 0xFFFFFFFF
        seed ^= (seed << 5) & 0xFFFFFFFF
        yield seed


def initialize_dictionary(size, seed=None):
    if seed is None:
        seed = 3758097560
    elif seed == 0:
        return bytearray(size)

    chars = b" \x000ei>to<ans\nr/."  # 16 most common chars in dataset

    def _gen_stream(xorshift32):
        for _ in range(size >> 3):
            value = next(xorshift32)
            yield chars[value & 0x0F]
            yield chars[value >> 4 & 0x0F]
            yield chars[value >> 8 & 0x0F]
            yield chars[value >> 12 & 0x0F]
            yield chars[value >> 16 & 0x0F]
            yield chars[value >> 20 & 0x0F]
            yield chars[value >> 24 & 0x0F]
            yield chars[value >> 28 & 0x0F]

    return bytearray(_gen_stream(_xorshift32(seed)))


def compute_min_pattern_size(window, literal):
    """Compute whether the minimum pattern length should be 2 or 3."""
    if window > 15 or window < 8:
        raise ValueError
    if literal == 5:
        return 2 + (window > 10)
    elif literal == 6:
        return 2 + (window > 12)
    elif literal == 7:
        return 2 + (window > 14)
    elif literal == 8:
        return 2
    else:
        raise ValueError


try:
    from .compressor_viper import Compressor, TextCompressor, compress
except ImportError:
    try:
        from ._c_compressor import Compressor, TextCompressor, compress
    except ImportError:
        try:
            from .compressor import Compressor, TextCompressor, compress
        except ImportError:
            pass

try:
    from .decompressor_viper import Decompressor, TextDecompressor, decompress
except ImportError:
    try:
        from ._c_decompressor import Decompressor, TextDecompressor, decompress
    except ImportError:
        try:
            from .decompressor import Decompressor, TextDecompressor, decompress
        except ImportError:
            pass


def open(f, mode="rb", **kwargs):
    if "r" in mode and "w" in mode:
        raise ValueError

    if "r" in mode:  # Decompressor
        if "b" in mode:
            return Decompressor(f, **kwargs)
        else:
            return TextDecompressor(f, **kwargs)
    elif "w" in mode:  # Compressor
        if "b" in mode:
            return Compressor(f, **kwargs)
        else:
            return TextCompressor(f, **kwargs)
    else:
        raise ValueError
